Require Import ModalLogic.syntax.

Require Import FunInd.
Require Import Coq.Logic.Classical_Prop.
Require Import Relations.Relation_Definitions.
Require Import Coq.Logic.Classical_Pred_Type.
Require Import Coq.Lists.List.
Require Import Coq.Sets.Ensembles.
Require Import Coq.Bool.Bool.
Require Import Coq.Bool.BoolEq.


Definition W := Type.

Definition Ra (A:Ensemble W) := Agents -> W -> W -> Prop.

Check Ra.

Inductive Path (A:Ensemble W) (w1 w2:W) (R:Ra A) : Prop:=
  |  empty : forall a, (R a (In W A w1) (In W A w2)) -> Path A w1 w2 R
  |  app: forall a v, (R a (In W A w1) (In W A v)) -> (Path A v w2 R) -> Path A w1 w2 R.

Compute Path.

Definition Ra_acyclic (A:Ensemble W) (R:Ra A) := forall (w:W), ~ Path A w w R.

Definition has_parent (A:Ensemble W) (R:Ra A) (w1:W) := exists a w, R a (In W A w) (In W A w1).

Definition is_root (A:Ensemble W) (R:Ra A) (w1:W) := exists a w3, R a (In W A w1) (In W A w3) /\ ~(has_parent A R w1).

Definition unique_root (A:Ensemble W) (R:Ra A) := exists w1, (is_root A R w1) /\ forall w2, ~(w2 = w1) -> ~(is_root A R w2).

Definition Frame (A:Ensemble W) : Type := (Ensemble W*Ra A).

Definition pi := W -> Props -> bool.

Definition Model (A:Ensemble W) : Type := 
  (Frame A* pi) %type.


Definition M_tree_like (A:Ensemble W) (M:Model A) := Ra_acyclic A (snd(fst(M))) /\ exists a, is_root A (snd(fst(M))) a /\ unique_root A (snd(fst(M))).

Fixpoint sat (A:Ensemble W) (M:Model A) (w:W) (f:formula): Prop :=
 let R := snd(fst(M)) in 
 let pi:= snd M in
  match f with 
  | ConstF tt => True
  | ConstF ff => False
  | PropsF p => if (pi w p) then True else False
  | Not g => ~(sat A M w g)
  | And g h => (sat A M w g) /\ (sat A M w h)
  | Or g h => (sat A M w g) \/ (sat A M w h)
  | Imp g h => ~(sat A M w g) \/ (sat A M w h)
  | Box a g => forall w':W, ((R a) (In W A w) (In W A w')) -> (sat A M w' g)
  | Diamond a g => exists w':W, ((R a) (In W A w) (In W A w')) /\ (sat A M w' g)
  end.


Definition locally_satisfatible (A:Ensemble W) (f:formula): Prop := exists (M:Model A), exists w:W, sat A M w f.

Definition globally_satisfatible (A:Ensemble W) (f:formula): Prop := exists (M:Model A), forall w:W, sat A M w f.

Definition eq_semantica (A:Ensemble W) (f g : formula) : Prop :=
  forall (M:Model A) (w:W), sat A M w f = sat A M w g.


Definition tree_locally_satisfatible (A:Ensemble W) (f:formula): Prop := exists (M:Model A), exists (w:W), M_tree_like A M /\  is_root A (snd(fst M)) w /\ sat A M w f.

Theorem sat_tree_and_model (A:Ensemble W) (f:formula) :
  tree_locally_satisfatible A f <->  locally_satisfatible A f.
Proof.
  split.
  (* The first part is easy: a tree-like model is a model *)
  + intro. unfold tree_locally_satisfatible in H. unfold locally_satisfatible.
    destruct H. exists x. destruct H. exists x0. destruct H. destruct H0. apply H1.
  (* For the second part, you need to construct a tree-like model M' from M, then prove that the constructed model indeed satisfies the formula *)
  + intro.  unfold locally_satisfatible in H. unfold tree_locally_satisfatible.
Admitted.

Definition pointed_model (A:Ensemble W) (w:W) (M:Model A) := (In W (fst (fst M)) w). (* tem que trabalhar na definição de modelo e/ou aqui também *)
